#################################################################################
##Implementation of ResNet-104 with triplet semi-hard metric learning############
#################################################################################


#Load necessary libraries
library(png)
library(gtools)
library("EBImage")
library(tensorflow)
library(keras)

#get directories of all images and masks
dirs=list.dirs(path = "C:/Users/Cell/Train", full.names = TRUE, recursive = FALSE)
dirs_images=paste(dirs,'/images', sep="")
dirs_masks=paste(dirs,'/masks', sep="")

#Function to read all one image masks and join them into a single image
join_masks<-function(masks_dir){
  masks <- list.files(path=masks_dir,pattern = "png",recursive = TRUE)
  
  img=readPNG(gsub(" ", "", paste(paste(masks_dir,"/"),masks[1]), fixed = TRUE))
  if(length(masks)>=2){
    for(i in 2:length(masks)){
      img=img+readPNG(gsub(" ", "", paste(paste(masks_dir,"/"),masks[i]), fixed = TRUE))
    }
  }
  
  return((img>0)*1)
}

#Read images and their masks into two arrays
#Images are read and resized to 256x256 format matrices
#In addition, RGB images are transformed to grey scale
all_train_images=array(0,c(670,256,256))
all_train_masks=array(0,c(670,256,256))
for(i in 1:670){
  image_dir <- list.files(path=dirs_images[i],pattern = "png",recursive = TRUE)
  img=readPNG(gsub(" ", "", paste(paste(dirs_images[i],"/"),image_dir), fixed = TRUE))     
  img=img[,,-dim(img)[3]]
  img=rowSums(img,dims=2)
  img=resize(img/max(img),256,256)
  get_mask=join_masks(dirs_masks[i])
  get_mask=resize(get_mask,256,256)
  all_train_images[i,,]=img
  all_train_masks[i,,]=get_mask
  cat(sprintf("image: %s\n", i))
}


#Create a random fold by selecting 536 images for traning
#and leaving 134 images for testing
set.seed(11)
idx=sample(seq(1,670,1),670,replace = F)
train_images=all_train_images[idx[1:536],,]
train_masks=all_train_masks[idx[1:536],,]
test_images=all_train_images[idx[537:670],,]
test_masks=all_train_masks[idx[537:670],,]

#######################################################
#Construct ResNet-104 network 
library(tensorflow)
X=tf$placeholder(tf$float32,shape(NULL,256L,256L,1L),name="X")

training = tf$placeholder(tf$bool)
conv11=tf$layers$batch_normalization(tf$layers$conv2d(X,filters=16L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)

#Each for cycle is creating new consecutive residual blocks
conv_1=conv11
for(iter in 1:2){
  input1=conv_1
  conv_1=tf$layers$batch_normalization(tf$layers$conv2d(input1,filters=8L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_1=tf$layers$batch_normalization(tf$layers$conv2d(conv_1,filters=8L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_1=tf$layers$batch_normalization(tf$layers$conv2d(conv_1,filters=16L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_1=tf$add(conv_1,input1)
}
conv_1=tf$layers$batch_normalization(tf$layers$conv2d(conv_1,filters=32L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
pool_1=tf$layers$max_pooling2d(conv_1,pool_size = 3L,strides = 2L,padding = "SAME")

conv_2=pool_1
for(iter in 1:2){
  input1=conv_2
  conv_2=tf$layers$batch_normalization(tf$layers$conv2d(input1,filters=16L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_2=tf$layers$batch_normalization(tf$layers$conv2d(conv_2,filters=16L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_2=tf$layers$batch_normalization(tf$layers$conv2d(conv_2,filters=32L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_2=tf$add(conv_2,input1)
}
conv_2=tf$layers$batch_normalization(tf$layers$conv2d(conv_2,filters=64L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
pool_2=tf$layers$max_pooling2d(conv_2,pool_size = 3L,strides = 2L,padding = "SAME")

conv_3=pool_2
for(iter in 1:4){
  input1=conv_3
  conv_3=tf$layers$batch_normalization(tf$layers$conv2d(input1,filters=32L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_3=tf$layers$batch_normalization(tf$layers$conv2d(conv_3,filters=32L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_3=tf$layers$batch_normalization(tf$layers$conv2d(conv_3,filters=64L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_3=tf$add(conv_3,input1)
}
conv_3=tf$layers$batch_normalization(tf$layers$conv2d(conv_3,filters=128L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
pool_3=tf$layers$max_pooling2d(conv_3,pool_size = 3L,strides = 2L,padding = "SAME")

conv_4=pool_3
for(iter in 1:4){
  input1=conv_4
  conv_4=tf$layers$batch_normalization(tf$layers$conv2d(input1,filters=64L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_4=tf$layers$batch_normalization(tf$layers$conv2d(conv_4,filters=64L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_4=tf$layers$batch_normalization(tf$layers$conv2d(conv_4,filters=128L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_4=tf$add(conv_4,input1)
}
conv_4=tf$layers$batch_normalization(tf$layers$conv2d(conv_4,filters=256L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
pool_4=tf$layers$max_pooling2d(conv_4,pool_size = 3L,strides = 2L,padding = "SAME")


conv_5=pool_4
for(iter in 1:6){
  input1=conv_5
  conv_5=tf$layers$batch_normalization(tf$layers$conv2d(input1,filters=128L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_5=tf$layers$batch_normalization(tf$layers$conv2d(conv_5,filters=128L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_5=tf$layers$batch_normalization(tf$layers$conv2d(conv_5,filters=256L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_5=tf$add(conv_5,input1)
}
conv_5=tf$layers$batch_normalization(tf$layers$conv2d(conv_5,filters=512L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
pool_5=tf$layers$max_pooling2d(conv_5,pool_size = 3L,strides = 2L,padding = "SAME")

conv_6=pool_5
for(iter in 1:6){
  input1=conv_6
  conv_6=tf$layers$batch_normalization(tf$layers$conv2d(input1,filters=256L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_6=tf$layers$batch_normalization(tf$layers$conv2d(conv_6,filters=256L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_6=tf$layers$batch_normalization(tf$layers$conv2d(conv_6,filters=512L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
  conv_6=tf$add(conv_6,input1)
}


#Decoder part: each upsampling is followed by 4 regular convolutional layers
up8=tf$layers$Conv2DTranspose(filters=512L,kernel_size = 2L,strides = 2L,padding = "SAME")(conv_6)
add_layers8=tf$concat(c(conv_5,up8),axis = 3L)
conv_8=tf$layers$batch_normalization(tf$layers$conv2d(add_layers8,filters=256L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_8=tf$layers$batch_normalization(tf$layers$conv2d(conv_8,filters=256L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_8=tf$layers$batch_normalization(tf$layers$conv2d(conv_8,filters=256L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_8=tf$layers$batch_normalization(tf$layers$conv2d(conv_8,filters=256L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)

up9=tf$layers$Conv2DTranspose(filters=256L,kernel_size = 2L,strides = 2L,padding = "SAME")(conv_8)
add_layers9=tf$concat(c(conv_4,up9),axis = 3L)
conv_9=tf$layers$batch_normalization(tf$layers$conv2d(add_layers9,filters=128L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_9=tf$layers$batch_normalization(tf$layers$conv2d(conv_9,filters=128L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_9=tf$layers$batch_normalization(tf$layers$conv2d(conv_9,filters=128L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_9=tf$layers$batch_normalization(tf$layers$conv2d(conv_9,filters=128L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)

up10=tf$layers$Conv2DTranspose(filters=128L,kernel_size = 2L,strides = 2L,padding = "SAME")(conv_9)
add_layers10=tf$concat(c(conv_3,up10),axis = 3L)
conv_10=tf$layers$batch_normalization(tf$layers$conv2d(add_layers10,filters=64L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_10=tf$layers$batch_normalization(tf$layers$conv2d(conv_10,filters=64L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_10=tf$layers$batch_normalization(tf$layers$conv2d(conv_10,filters=64L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_10=tf$layers$batch_normalization(tf$layers$conv2d(conv_10,filters=64L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)

up11=tf$layers$Conv2DTranspose(filters=64L,kernel_size = 2L,strides = 2L,padding = "SAME")(conv_10)
add_layers11=tf$concat(c(conv_2,up11),axis = 3L)
conv_11=tf$layers$batch_normalization(tf$layers$conv2d(add_layers11,filters=32L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_11=tf$layers$batch_normalization(tf$layers$conv2d(conv_11,filters=32L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_11=tf$layers$batch_normalization(tf$layers$conv2d(conv_11,filters=32L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_11=tf$layers$batch_normalization(tf$layers$conv2d(conv_11,filters=32L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)

up12=tf$layers$Conv2DTranspose(filters=32L,kernel_size = 2L,strides = 2L,padding = "SAME")(conv_11)
add_layers12=tf$concat(c(conv_1,up12),axis = 3L)
conv_12=tf$layers$batch_normalization(tf$layers$conv2d(add_layers12,filters=16L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_12=tf$layers$batch_normalization(tf$layers$conv2d(conv_12,filters=16L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_12=tf$layers$batch_normalization(tf$layers$conv2d(conv_12,filters=16L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)
conv_12=tf$layers$batch_normalization(tf$layers$conv2d(conv_12,filters=16L,kernel_size = 3L,strides = 1L,padding = "SAME",activation = tf$nn$leaky_relu,kernel_initializer = tf$orthogonal_initializer),training=training)


conv_14=tf$layers$conv2d(conv_12,filters=1L,kernel_size = 1L,strides = 1L,padding = "SAME",activation = NULL,kernel_initializer = tf$orthogonal_initializer)
conv14_sig=tf$nn$sigmoid(conv_14)

#Creat a space in the tensorflow graph mask input
cell_mask=tf$placeholder(tf$float32,shape(NULL,256L,256L))
#Flatten 12th layer as well as last layer and mask tensor
last_flat=tf$reshape(conv_12,c(4L*256L*256L,16L))
pred_flat=tf$squeeze(tf$reshape(conv14_sig,c(4L*256L*256L,1L)))
mask_flat=tf$squeeze(tf$reshape(cell_mask,c(4L*256L*256L,1L)))

##########Calculation of triplet semi-hard metric###########
#Create a place to transfer into a tensorflow graph indices of
#400 pixes - these pixels are selected in the training loop below
#These pixels will be used to calculated triplet semi-hard loss function value
ridxs=tf$placeholder(tf$int64,c(400L))
#Collect those pixels from 12th layer together 
#with their label from the masks
last_flat_sub = tf$gather(last_flat,tf$cast(ridxs,dtype =tf$int64))
mask_flat_sub = tf$gather(mask_flat,tf$cast(ridxs,dtype =tf$int64))
#Perform L2 normalization and input into a triplet semi-hard metric function
h_flat1_norm=last_flat_sub/(tf$norm(last_flat_sub,ord="euclidean",axis=1L,keepdims=1L)+0.0001)
triplet_loss=1*tf$contrib$losses$metric_learning$triplet_semihard_loss (labels=mask_flat_sub,embeddings =h_flat1_norm,margin=1.0 )
#triplet_loss=10/400*tf$contrib$losses$metric_learning$lifted_struct_loss(labels=mask_flat_sub,embeddings =last_flat_sub,margin=1.0 )

###############Calculation of dice loss function##############
y_true_f = tf$layers$flatten(cell_mask)
y_pred_f = tf$layers$flatten(conv14_sig)
intersection = tf$reduce_sum(y_true_f * y_pred_f,1L)
dice_loss=tf$reduce_mean(1-(2*intersection+1.0)/(1.0+tf$reduce_sum(y_true_f,1L)+tf$reduce_sum(y_pred_f,1L)))
###############Calculation of cross-entropy loss function#####
logits = tf$squeeze(conv_14)
learning_coef = tf$placeholder(tf$float32)
reconstruction_loss = tf$reduce_mean(tf$nn$sigmoid_cross_entropy_with_logits(labels = cell_mask, logits=logits))

#Adding all three metrics
comb_loss=triplet_loss+dice_loss+learning_coef*reconstruction_loss
#Add to the graph additional operators, which are needed 
#because of batchnormalization layers
extra_update_ops =tf$get_collection(tf$GraphKeys$UPDATE_OPS)
#Setting up optimizer
learning_rate = tf$placeholder(tf$float32)
optimizer1 <- tf$train$AdamOptimizer(learning_rate,epsilon = 0.01)
train_step1 <- optimizer1$minimize(comb_loss)

#Setting up the session
init <- tf$global_variables_initializer()
gpu_options <- tf$GPUOptions(allow_growth = TRUE)
config <- tf$ConfigProto(gpu_options = gpu_options)
sess <- tf$Session(config = config)
sess$run(init)

###########################################################
#Network training, testing and visualization###############


cost_train<-matrix(0,1,10000)
cost_test<-matrix(0,1,10000)
validation_dice=matrix(0,100,1)
rec_train<-matrix(0,1,10000)
rec_test<-matrix(0,1,10000)
par(mfrow=c(3,2),tcl=-0.5, family="serif", mai=c(0.3,0.3,0.3,0.3))
count=0
miniBatch=4
N=dim(train_images)[1]
imTest=array(0,c(4,256,256,1))
maskTest=array(0,c(4,256,256))
imTrain=array(0,c(miniBatch,256,256,1))
maskTrain=array(0,c(miniBatch,256,256))
prd=matrix(0,miniBatch,256*256)
collect=0
collect_test=0
collect_validation=0
for(e in 1:1000){
  #mix idices of training set
  m=sample(seq(1,N,1),N,replace=F)
  #Cycle over a one epoch
  for (i in seq(1,length(m)-miniBatch,1*miniBatch)){
    toIdx=i+miniBatch-1
    takeTrain=m[i:toIdx]
    imTrain[,,,1]=train_images[takeTrain,,]
    maskTrain[,,]=train_masks[takeTrain,,]
    
    for(n in 1:miniBatch){
      #To increase the sample size, rotations on images and masks are performed
      rot_angle=sample(seq(0,360,1),1)
      img=imTrain[n,,,1]
      img_mask=maskTrain[n,,]
      
      img=rotate(img,rot_angle)
      img_mask=rotate(img_mask,rot_angle)
      
      img=resize(img,256,256)
      img_mask=resize(img_mask,256,256)
      imTrain[n,,,1]=img
      maskTrain[n,,]=img_mask
    }
    #Select 200 pixels from background and 200 pixels from nucleus
    #This is required for trilet semi-hard metric calculations
    tmp_idx=sess$run(mask_flat,feed_dict = dict(X = imTrain, cell_mask= maskTrain,training=FALSE))
    idx0=which(tmp_idx==0)
    idx1=which(tmp_idx>0)
    idx0=sample(idx0,200,replace = F)
    idx1=sample(idx1,200,replace=F)
    #Feed the batch of images, masks and selected pixels to the graph and perform on update of parameters
    sess$run(c(train_step1,extra_update_ops),feed_dict = dict(X= imTrain,cell_mask= maskTrain, training=TRUE,learning_rate=0.0005,learning_coef=1.0,ridxs=c(idx0,idx1)))      
    #Calculate dice and triplet losses
    collect=collect+1
    cost_train[collect]=sess$run(dice_loss,feed_dict = dict(X = imTrain, cell_mask= maskTrain,training=FALSE,ridxs=c(idx0,idx1)))
    rec_train[collect]=sess$run(triplet_loss,feed_dict = dict(X = imTrain, cell_mask= maskTrain,training=FALSE,ridxs=c(idx0,idx1)))
    
    #Track network performance on testing set
    if(collect%%50==0){
      collect_test=collect_test+1
      mt=sample(seq(1,dim(test_images)[1],1),4,replace=F)
      imTest[,,,1]=test_images[mt,,]
      maskTest[,,]=test_masks[mt,,]
      
      tmp_idx=sess$run(mask_flat,feed_dict = dict(X = imTest, cell_mask= maskTest,training=FALSE))
      idx0=which(tmp_idx==0)
      idx1=which(tmp_idx>0)
      idx0=sample(idx0,200,replace = F)
      idx1=sample(idx1,200,replace=F)
      
      rec_test[collect_test]=sess$run(triplet_loss,feed_dict = dict(X = imTest, cell_mask= maskTest,training=FALSE,ridxs=c(idx0,idx1)))
      cost_test[collect_test]=sess$run(dice_loss,feed_dict = dict(X = imTest, cell_mask= maskTest,training=FALSE,ridxs=c(idx0,idx1)))
      cat(sprintf("cost_test %s\n", cost_test[collect_test]))
      plot( cost_train[1:collect],ylab="dice train")
      plot(cost_test[1:collect_test],ylab="dice test")
      plot( rec_train[1:collect],ylab="reconstruct train")
      plot(rec_test[1:collect_test],ylab="reconstruct test")
      display(maskTest[1,,],method="raster")
      pred_mask=sess$run(conv14_sig,feed_dict = dict(X = imTest, cell_mask= maskTest,training=FALSE,ridxs=c(idx0,idx1)))
      display(pred_mask[1,,,1],method="raster")
      
    }
    
  }
  cat(sprintf("epoch %s\n", e))
}


